@@ -9,7 +9,9 @@
  * SPDX-License-Identifier: MIT
  */
 
+#include <sys/stat.h>
 #include <sys/sysinfo.h>
+#include <sys/sysmacros.h>
 
 #include "util/disk_cache.h"
 #include "git_sha1.h"
@@ -34,7 +36,6 @@
 #include "genxml/gen_macros.h"
 
 #define ARM_VENDOR_ID        0x13b5
-#define MAX_VIEWPORTS        1
 #define MAX_PUSH_DESCRIPTORS 32
 /* We reserve one ubo for push constant, one for sysvals and one per-set for the
  * descriptor metadata  */
@@ -42,6 +43,87 @@
 #define MAX_INLINE_UNIFORM_BLOCK_DESCRIPTORS 32 - RESERVED_UBO_COUNT
 #define MAX_INLINE_UNIFORM_BLOCK_SIZE        (1 << 16)
 
+static VkResult
+create_kmod_dev(struct panvk_physical_device *device,
+                const struct panvk_instance *instance, drmDevicePtr drm_device)
+{
+   const char *path = drm_device->nodes[DRM_NODE_RENDER];
+   drmVersionPtr version;
+   int fd;
+
+   fd = open(path, O_RDWR | O_CLOEXEC);
+   if (fd < 0) {
+      return panvk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
+                          "failed to open device %s", path);
+   }
+
+   version = drmGetVersion(fd);
+   if (!version) {
+      close(fd);
+      return panvk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
+                          "failed to query kernel driver version for device %s",
+                          path);
+   }
+
+   if (strcmp(version->name, "panfrost") && strcmp(version->name, "panthor")) {
+      drmFreeVersion(version);
+      close(fd);
+      return panvk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
+                          "device %s does not use the panfrost kernel driver",
+                          path);
+   }
+
+   drmFreeVersion(version);
+
+   if (!getenv("PAN_I_WANT_A_BROKEN_VULKAN_DRIVER")) {
+      close(fd);
+      return panvk_errorf(
+         instance, VK_ERROR_INCOMPATIBLE_DRIVER,
+         "WARNING: panvk is not a conformant vulkan implementation, "
+         "pass PAN_I_WANT_A_BROKEN_VULKAN_DRIVER=1 if you know what you're doing.");
+   }
+
+   if (instance->debug_flags & PANVK_DEBUG_STARTUP)
+      vk_logi(VK_LOG_NO_OBJS(instance), "Found compatible device '%s'.", path);
+
+   device->kmod.dev = pan_kmod_dev_create(fd, PAN_KMOD_DEV_FLAG_OWNS_FD,
+                                          &instance->kmod.allocator);
+
+   if (!device->kmod.dev) {
+      close(fd);
+      return panvk_errorf(instance, VK_ERROR_OUT_OF_HOST_MEMORY,
+                          "cannot create device");
+   }
+
+   return VK_SUCCESS;
+}
+
+static VkResult
+get_drm_device_ids(struct panvk_physical_device *device,
+                   const struct panvk_instance *instance,
+                   drmDevicePtr drm_device)
+{
+   struct stat st;
+
+   if (stat(drm_device->nodes[DRM_NODE_RENDER], &st)) {
+      return vk_errorf(instance, VK_ERROR_INITIALIZATION_FAILED,
+                       "failed to query render node stat");
+   }
+
+   device->drm.render_rdev = st.st_rdev;
+
+   if (drm_device->available_nodes & (1 << DRM_NODE_PRIMARY)) {
+      if (stat(drm_device->nodes[DRM_NODE_PRIMARY], &st)) {
+         return vk_errorf(instance, VK_ERROR_INITIALIZATION_FAILED,
+                          "failed to query primary node stat");
+      }
+
+      device->drm.primary_rdev = st.st_rdev;
+   }
+
+   return VK_SUCCESS;
+}
+
 static int
 get_cache_uuid(uint16_t family, void *uuid)
 {
@@ -58,20 +140,75 @@ get_cache_uuid(uint16_t family, void *uuid)
    return 0;
 }
 
+static VkResult
+get_device_sync_types(struct panvk_physical_device *device,
+                      const struct panvk_instance *instance)
+{
+   const unsigned arch = pan_arch(device->kmod.props.gpu_prod_id);
+   uint32_t sync_type_count = 0;
+
+   device->drm_syncobj_type = vk_drm_syncobj_get_type(device->kmod.dev->fd);
+   if (!device->drm_syncobj_type.features) {
+      return vk_errorf(instance, VK_ERROR_INITIALIZATION_FAILED,
+                       "failed to query syncobj features");
+   }
+
+   device->sync_types[sync_type_count++] = &device->drm_syncobj_type;
+
+   if (arch >= 10) {
+      assert(device->drm_syncobj_type.features & VK_SYNC_FEATURE_TIMELINE);
+   } else {
+      /* We don't support timelines in the uAPI yet and we don't want it getting
+       * suddenly turned on by vk_drm_syncobj_get_type() without us adding panvk
+       * code for it first.
+       */
+      device->drm_syncobj_type.features &= ~VK_SYNC_FEATURE_TIMELINE;
+
+      /* vk_sync_timeline requires VK_SYNC_FEATURE_GPU_MULTI_WAIT.  Panfrost
+       * waits on the underlying dma-fences and supports the feature.
+       */
+      device->drm_syncobj_type.features |= VK_SYNC_FEATURE_GPU_MULTI_WAIT;
+
+      device->sync_timeline_type =
+         vk_sync_timeline_get_type(&device->drm_syncobj_type);
+      device->sync_types[sync_type_count++] = &device->sync_timeline_type.sync;
+   }
+
+   assert(sync_type_count < ARRAY_SIZE(device->sync_types));
+   device->sync_types[sync_type_count] = NULL;
+
+   return VK_SUCCESS;
+}
+
 static void
 get_device_extensions(const struct panvk_physical_device *device,
                       struct vk_device_extension_table *ext)
 {
    *ext = (struct vk_device_extension_table){
+      .KHR_16bit_storage = true,
+      .KHR_bind_memory2 = true,
       .KHR_buffer_device_address = true,
       .KHR_copy_commands2 = true,
+      .KHR_create_renderpass2 = true,
       .KHR_device_group = true,
       .KHR_descriptor_update_template = true,
       .KHR_driver_properties = true,
+      .KHR_dynamic_rendering = true,
+      .KHR_external_fence = true,
+      .KHR_external_fence_fd = true,
+      .KHR_external_memory = true,
+      .KHR_external_memory_fd = true,
+      .KHR_external_semaphore = true,
+      .KHR_external_semaphore_fd = true,
+      .KHR_get_memory_requirements2 = true,
+      .KHR_image_format_list = true,
+      .KHR_maintenance1 = true,
+      .KHR_maintenance2 = true,
       .KHR_maintenance3 = true,
       .KHR_pipeline_executable_properties = true,
       .KHR_pipeline_library = true,
       .KHR_push_descriptor = true,
+      .KHR_relaxed_block_layout = true,
       .KHR_sampler_mirror_clamp_to_edge = true,
       .KHR_shader_expect_assume = true,
       .KHR_storage_buffer_storage_class = true,
@@ -79,16 +216,24 @@ get_device_extensions(const struct panvk_physical_device *device,
       .KHR_swapchain = true,
 #endif
       .KHR_synchronization2 = true,
+      .KHR_timeline_semaphore = true,
       .KHR_variable_pointers = true,
       .EXT_buffer_device_address = true,
       .EXT_custom_border_color = true,
+      .EXT_depth_clip_enable = true,
+      .EXT_external_memory_dma_buf = true,
       .EXT_graphics_pipeline_library = true,
+      .EXT_image_drm_format_modifier = true,
       .EXT_index_type_uint8 = true,
+      .EXT_physical_device_drm = true,
       .EXT_pipeline_creation_cache_control = true,
       .EXT_pipeline_creation_feedback = true,
       .EXT_private_data = true,
       .EXT_shader_module_identifier = true,
       .EXT_vertex_attribute_divisor = true,
+      .GOOGLE_decorate_string = true,
+      .GOOGLE_hlsl_functionality1 = true,
+      .GOOGLE_user_type = true,
    };
 }
 
@@ -116,7 +261,7 @@ get_features(const struct panvk_physical_device *device,
 
       /* Vulkan 1.1 */
       .storageBuffer16BitAccess = false,
-      .uniformAndStorageBuffer16BitAccess = false,
+      .uniformAndStorageBuffer16BitAccess = true,
       .storagePushConstant16 = false,
       .storageInputOutput16 = false,
       .multiview = false,
@@ -168,7 +313,7 @@ get_features(const struct panvk_physical_device *device,
       .shaderSubgroupExtendedTypes = false,
       .separateDepthStencilLayouts = false,
       .hostQueryReset = false,
-      .timelineSemaphore = false,
+      .timelineSemaphore = true,
       .bufferDeviceAddress = true,
       .bufferDeviceAddressCaptureReplay = false,
       .bufferDeviceAddressMultiDevice = false,
@@ -192,7 +337,7 @@ get_features(const struct panvk_physical_device *device,
       .synchronization2 = true,
       .textureCompressionASTC_HDR = false,
       .shaderZeroInitializeWorkgroupMemory = false,
-      .dynamicRendering = false,
+      .dynamicRendering = true,
       .shaderIntegerDotProduct = false,
       .maintenance4 = false,
 
@@ -246,7 +391,7 @@ get_device_properties(const struct panvk_instance *instance,
    uint64_t os_page_size = 4096;
    os_get_page_size(&os_page_size);
 
-   ASSERTED unsigned arch = pan_arch(device->kmod.props.gpu_prod_id);
+   unsigned arch = pan_arch(device->kmod.props.gpu_prod_id);
 
    /* Ensure that the max threads count per workgroup is valid for Bifrost */
    assert(arch > 8 || device->kmod.props.max_threads_per_wg <= 1024);
@@ -295,14 +440,18 @@ get_device_properties(const struct panvk_instance *instance,
       .bufferImageGranularity = 64,
       /* Sparse binding not supported yet. */
       .sparseAddressSpaceSize = 0,
-      /* Software limit. Pick the minimum required by Vulkan, because Bifrost
-       * GPUs don't have unified descriptor tables, which forces us to
-       * agregatte all descriptors from all sets and dispatch them to per-type
-       * descriptor tables emitted at draw/dispatch time.
-       * The more sets we support the more copies we are likely to have to do
-       * at draw time.
+      /* On Bifrost, this is a software limit. We pick the minimum required by
+       * Vulkan, because Bifrost GPUs don't have unified descriptor tables,
+       * which forces us to agregatte all descriptors from all sets and dispatch
+       * them to per-type descriptor tables emitted at draw/dispatch time. The
+       * more sets we support the more copies we are likely to have to do at
+       * draw time.
+       *
+       * Valhall has native support for descriptor sets, and allows a maximum
+       * of 16 sets, but we reserve one for our internal use, so we have 15
+       * left.
        */
-      .maxBoundDescriptorSets = 4,
+      .maxBoundDescriptorSets = arch <= 7 ? 4 : 15,
       /* MALI_RENDERER_STATE::sampler_count is 16-bit. */
       .maxDescriptorSetSamplers = UINT16_MAX,
       /* MALI_RENDERER_STATE::uniform_buffer_count is 8-bit. We reserve 32 slots
@@ -407,7 +556,7 @@ get_device_properties(const struct panvk_instance *instance,
       .maxDrawIndexedIndexValue = UINT32_MAX,
       /* Make it one for now. */
       .maxDrawIndirectCount = 1,
-      .maxSamplerLodBias = 255,
+      .maxSamplerLodBias = (float)INT16_MAX / 256.0f,
       .maxSamplerAnisotropy = 16,
       .maxViewports = 1,
       /* Same as the framebuffer limit. */
@@ -543,7 +692,6 @@ get_device_properties(const struct panvk_instance *instance,
       /* XXX: VK_EXT_sampler_filter_minmax */
       .filterMinmaxSingleComponentFormats = false,
       .filterMinmaxImageComponentMapping = false,
-      /* XXX: VK_KHR_timeline_semaphore */
       .maxTimelineSemaphoreValueDifference = INT64_MAX,
       .framebufferIntegerColorSampleCounts = sample_counts,
 
@@ -618,6 +766,18 @@ get_device_properties(const struct panvk_instance *instance,
    snprintf(properties->driverInfo, VK_MAX_DRIVER_INFO_SIZE,
             "Mesa " PACKAGE_VERSION MESA_GIT_SHA1);
 
+   /* VK_EXT_physical_device_drm */
+   if (device->drm.primary_rdev) {
+      properties->drmHasPrimary = true;
+      properties->drmPrimaryMajor = major(device->drm.primary_rdev);
+      properties->drmPrimaryMinor = minor(device->drm.primary_rdev);
+   }
+   if (device->drm.render_rdev) {
+      properties->drmHasRender = true;
+      properties->drmRenderMajor = major(device->drm.render_rdev);
+      properties->drmRenderMinor = minor(device->drm.render_rdev);
+   }
+
    /* VK_EXT_shader_module_identifier */
    STATIC_ASSERT(sizeof(vk_shaderModuleIdentifierAlgorithmUUID) ==
                  sizeof(properties->shaderModuleIdentifierAlgorithmUUID));
@@ -632,8 +792,6 @@ panvk_physical_device_finish(struct panvk_physical_device *device)
    panvk_wsi_finish(device);
 
    pan_kmod_dev_destroy(device->kmod.dev);
-   if (device->master_fd != -1)
-      close(device->master_fd);
 
    vk_physical_device_finish(&device->vk);
 }
@@ -643,75 +801,35 @@ panvk_physical_device_init(struct panvk_physical_device *device,
                            struct panvk_instance *instance,
                            drmDevicePtr drm_device)
 {
-   const char *path = drm_device->nodes[DRM_NODE_RENDER];
-   VkResult result = VK_SUCCESS;
-   drmVersionPtr version;
-   int fd;
-   int master_fd = -1;
+   VkResult result;
 
-   if (!getenv("PAN_I_WANT_A_BROKEN_VULKAN_DRIVER")) {
-      return vk_errorf(
-         instance, VK_ERROR_INCOMPATIBLE_DRIVER,
-         "WARNING: panvk is not a conformant vulkan implementation, "
-         "pass PAN_I_WANT_A_BROKEN_VULKAN_DRIVER=1 if you know what you're doing.");
-   }
-
-   fd = open(path, O_RDWR | O_CLOEXEC);
-   if (fd < 0) {
-      return vk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
-                       "failed to open device %s", path);
-   }
-
-   version = drmGetVersion(fd);
-   if (!version) {
-      close(fd);
-      return vk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
-                       "failed to query kernel driver version for device %s",
-                       path);
-   }
-
-   if (strcmp(version->name, "panfrost")) {
-      drmFreeVersion(version);
-      close(fd);
-      return vk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
-                       "device %s does not use the panfrost kernel driver",
-                       path);
-   }
-
-   drmFreeVersion(version);
-
-   if (instance->debug_flags & PANVK_DEBUG_STARTUP)
-      vk_logi(VK_LOG_NO_OBJS(instance), "Found compatible device '%s'.", path);
-
-   device->kmod.dev = pan_kmod_dev_create(fd, PAN_KMOD_DEV_FLAG_OWNS_FD,
-                                          &instance->kmod.allocator);
-
-   if (!device->kmod.dev) {
-      result = vk_errorf(instance, panvk_errno_to_vk_error(), "cannot create device");
-      goto fail;
-   }
+   result = create_kmod_dev(device, instance, drm_device);
+   if (result != VK_SUCCESS)
+      return result;
 
    pan_kmod_dev_query_props(device->kmod.dev, &device->kmod.props);
 
+   device->model = panfrost_get_model(device->kmod.props.gpu_prod_id,
+                                      device->kmod.props.gpu_variant);
+
    unsigned arch = pan_arch(device->kmod.props.gpu_prod_id);
 
-   if (arch <= 5 || arch >= 8) {
-      result = vk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
-                         "%s not supported", device->model->name);
+   switch (arch) {
+   case 6:
+   case 7:
+   case 10:
+      break;
+
+   default:
+      result = panvk_errorf(instance, VK_ERROR_INCOMPATIBLE_DRIVER,
+                            "%s not supported", device->model->name);
       goto fail;
    }
 
-   if (instance->vk.enabled_extensions.KHR_display) {
-      master_fd = open(drm_device->nodes[DRM_NODE_PRIMARY], O_RDWR | O_CLOEXEC);
-      if (master_fd >= 0) {
-         /* TODO: free master_fd is accel is not working? */
-      }
-   }
+   result = get_drm_device_ids(device, instance, drm_device);
+   if (result != VK_SUCCESS)
+      goto fail;
 
-   device->master_fd = master_fd;
-
-   device->model = panfrost_get_model(device->kmod.props.gpu_prod_id,
-                                      device->kmod.props.gpu_variant);
    device->formats.all = panfrost_format_table(arch);
    device->formats.blendable = panfrost_blendable_format_table(arch);
 
@@ -719,19 +837,16 @@ panvk_physical_device_init(struct panvk_physical_device *device,
    sprintf(device->name, "%s", device->model->name);
 
    if (get_cache_uuid(device->kmod.props.gpu_prod_id, device->cache_uuid)) {
-      result = vk_errorf(instance, VK_ERROR_INITIALIZATION_FAILED,
-                         "cannot generate UUID");
+      result = panvk_errorf(instance, VK_ERROR_INITIALIZATION_FAILED,
+                            "cannot generate UUID");
       goto fail;
    }
 
-   vk_warn_non_conformant_implementation("panvk");
+   result = get_device_sync_types(device, instance);
+   if (result != VK_SUCCESS)
+      goto fail;
 
-   device->drm_syncobj_type = vk_drm_syncobj_get_type(device->kmod.dev->fd);
-   /* We don't support timelines in the uAPI yet and we don't want it getting
-    * suddenly turned on by vk_drm_syncobj_get_type() without us adding panvk
-    * code for it first.
-    */
-   device->drm_syncobj_type.features &= ~VK_SYNC_FEATURE_TIMELINE;
+   vk_warn_non_conformant_implementation("panvk");
 
    struct vk_device_extension_table supported_extensions;
    get_device_extensions(device, &supported_extensions);
@@ -752,20 +867,14 @@ panvk_physical_device_init(struct panvk_physical_device *device,
                                     &supported_extensions, &supported_features,
                                     &properties, &dispatch_table);
 
-   if (result != VK_SUCCESS) {
-      vk_error(instance, result);
+   if (result != VK_SUCCESS)
       goto fail;
-   }
 
-   device->sync_types[0] = &device->drm_syncobj_type;
-   device->sync_types[1] = NULL;
    device->vk.supported_sync_types = device->sync_types;
 
    result = panvk_wsi_init(device);
-   if (result != VK_SUCCESS) {
-      vk_error(instance, result);
+   if (result != VK_SUCCESS)
       goto fail;
-   }
 
    return VK_SUCCESS;
 
@@ -773,13 +882,8 @@ fail:
    if (device->vk.instance)
       vk_physical_device_finish(&device->vk);
 
-   if (device->kmod.dev)
-      pan_kmod_dev_destroy(device->kmod.dev);
+   pan_kmod_dev_destroy(device->kmod.dev);
 
-   if (fd != -1)
-      close(fd);
-   if (master_fd != -1)
-      close(master_fd);
    return result;
 }
 
@@ -842,43 +946,6 @@ panvk_GetPhysicalDeviceMemoryProperties2(
    };
 }
 
-VKAPI_ATTR void VKAPI_CALL
-panvk_GetPhysicalDeviceExternalSemaphoreProperties(
-   VkPhysicalDevice physicalDevice,
-   const VkPhysicalDeviceExternalSemaphoreInfo *pExternalSemaphoreInfo,
-   VkExternalSemaphoreProperties *pExternalSemaphoreProperties)
-{
-   if ((pExternalSemaphoreInfo->handleType ==
-           VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_FD_BIT ||
-        pExternalSemaphoreInfo->handleType ==
-           VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_SYNC_FD_BIT)) {
-      pExternalSemaphoreProperties->exportFromImportedHandleTypes =
-         VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_FD_BIT |
-         VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_SYNC_FD_BIT;
-      pExternalSemaphoreProperties->compatibleHandleTypes =
-         VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_OPAQUE_FD_BIT |
-         VK_EXTERNAL_SEMAPHORE_HANDLE_TYPE_SYNC_FD_BIT;
-      pExternalSemaphoreProperties->externalSemaphoreFeatures =
-         VK_EXTERNAL_SEMAPHORE_FEATURE_EXPORTABLE_BIT |
-         VK_EXTERNAL_SEMAPHORE_FEATURE_IMPORTABLE_BIT;
-   } else {
-      pExternalSemaphoreProperties->exportFromImportedHandleTypes = 0;
-      pExternalSemaphoreProperties->compatibleHandleTypes = 0;
-      pExternalSemaphoreProperties->externalSemaphoreFeatures = 0;
-   }
-}
-
-VKAPI_ATTR void VKAPI_CALL
-panvk_GetPhysicalDeviceExternalFenceProperties(
-   VkPhysicalDevice physicalDevice,
-   const VkPhysicalDeviceExternalFenceInfo *pExternalFenceInfo,
-   VkExternalFenceProperties *pExternalFenceProperties)
-{
-   pExternalFenceProperties->exportFromImportedHandleTypes = 0;
-   pExternalFenceProperties->compatibleHandleTypes = 0;
-   pExternalFenceProperties->externalFenceFeatures = 0;
-}
-
 #define DEVICE_PER_ARCH_FUNCS(_ver)                                            \
    VkResult panvk_v##_ver##_create_device(                                     \
       struct panvk_physical_device *physical_device,                           \
@@ -890,6 +957,7 @@ panvk_GetPhysicalDeviceExternalFenceProperties(
 
 DEVICE_PER_ARCH_FUNCS(6);
 DEVICE_PER_ARCH_FUNCS(7);
+DEVICE_PER_ARCH_FUNCS(10);
 
 VKAPI_ATTR VkResult VKAPI_CALL
 panvk_CreateDevice(VkPhysicalDevice physicalDevice,
@@ -917,31 +985,59 @@ panvk_DestroyDevice(VkDevice _device, const VkAllocationCallbacks *pAllocator)
    panvk_arch_dispatch(arch, destroy_device, device, pAllocator);
 }
 
+static bool
+format_is_supported(struct panvk_physical_device *physical_device,
+                    const struct panfrost_format fmt,
+                    enum pipe_format pfmt)
+{
+   /* If the format ID is zero, it's not supported. */
+   if (!fmt.hw)
+      return false;
+
+   /* Compressed formats (ID < 32) are optional. We need to check against
+    * the supported formats reported by the GPU. */
+   if (util_format_is_compressed(pfmt)) {
+      uint32_t supported_compr_fmts =
+         panfrost_query_compressed_formats(&physical_device->kmod.props);
+
+      if (!(BITFIELD_BIT(fmt.texfeat_bit) & supported_compr_fmts))
+         return false;
+   }
+
+   return true;
+}
+
 static void
 get_format_properties(struct panvk_physical_device *physical_device,
                       VkFormat format, VkFormatProperties *out_properties)
 {
    VkFormatFeatureFlags tex = 0, buffer = 0;
    enum pipe_format pfmt = vk_format_to_pipe_format(format);
+   unsigned arch = pan_arch(physical_device->kmod.props.gpu_prod_id);
+
+   /* FIXME: Valhall doesn't support interleaved D32_S8X24. Implement it as
+    * a multi-plane format, and we probably want to switch Bifrost to this
+    * layout too, since:
+    * - it's more cache-friendly (you load more samples on a cache-line if you don't
+    *   have those 24 dummy bits)
+    * - it takes less memory (you don't lose those 24bits per texel)
+    * - we can use AFBC
+    */
+   if (arch >= 9 && format == VK_FORMAT_D32_SFLOAT_S8_UINT)
+      goto end;
+
+   if (pfmt == PIPE_FORMAT_NONE)
+      goto end;
+
    const struct panfrost_format fmt = physical_device->formats.all[pfmt];
 
-   if (!pfmt || !fmt.hw)
+   if (!format_is_supported(physical_device, fmt, pfmt))
       goto end;
 
    /* 3byte formats are not supported by the buffer <-> image copy helpers. */
    if (util_format_get_blocksize(pfmt) == 3)
       goto end;
 
-   /* We don't support compressed formats yet: this is causing trouble when
-    * doing a vkCmdCopyImage() between a compressed and a non-compressed format
-    * on a tiled/AFBC resource.
-    */
-   if (util_format_is_compressed(pfmt))
-      goto end;
-
-   buffer |=
-      VK_FORMAT_FEATURE_TRANSFER_SRC_BIT | VK_FORMAT_FEATURE_TRANSFER_DST_BIT;
-
    /* Reject sRGB formats (see
     * https://github.com/KhronosGroup/Vulkan-Docs/issues/2214).
     */
@@ -959,21 +1055,33 @@ get_format_properties(struct panvk_physical_device *physical_device,
       if (!util_format_is_scaled(pfmt) && !util_format_is_pure_integer(pfmt))
          tex |= VK_FORMAT_FEATURE_SAMPLED_IMAGE_FILTER_LINEAR_BIT;
 
-      buffer |= VK_FORMAT_FEATURE_UNIFORM_TEXEL_BUFFER_BIT;
+      if (!util_format_is_depth_or_stencil(pfmt))
+         buffer |= VK_FORMAT_FEATURE_UNIFORM_TEXEL_BUFFER_BIT;
 
       tex |= VK_FORMAT_FEATURE_BLIT_SRC_BIT;
    }
 
-   /* SNORM rendering isn't working yet, disable */
-   if (fmt.bind & PAN_BIND_RENDER_TARGET && !util_format_is_snorm(pfmt)) {
-      tex |= VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT |
-             VK_FORMAT_FEATURE_BLIT_DST_BIT;
-
+   if (fmt.bind & PAN_BIND_RENDER_TARGET) {
+      tex |= VK_FORMAT_FEATURE_BLIT_DST_BIT;
       tex |= VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT;
-      buffer |= VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_BIT;
 
-      /* Can always blend via blend shaders */
-      tex |= VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT;
+      /* SNORM rendering isn't working yet (nir_lower_blend bugs), disable for
+       * now.
+       *
+       * XXX: Enable once fixed.
+       */
+      if (!util_format_is_snorm(pfmt)) {
+         tex |= VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT;
+         tex |= VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BLEND_BIT;
+      }
+
+      if (!util_format_is_depth_and_stencil(pfmt))
+         buffer |= VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_BIT;
+   }
+
+   if (pfmt == PIPE_FORMAT_R32_UINT || pfmt == PIPE_FORMAT_R32_SINT) {
+      buffer |= VK_FORMAT_FEATURE_STORAGE_TEXEL_BUFFER_ATOMIC_BIT;
+      tex |= VK_FORMAT_FEATURE_STORAGE_IMAGE_ATOMIC_BIT;
    }
 
    if (fmt.bind & PAN_BIND_DEPTH_STENCIL)
@@ -985,16 +1093,6 @@ end:
    out_properties->bufferFeatures = buffer;
 }
 
-VKAPI_ATTR void VKAPI_CALL
-panvk_GetPhysicalDeviceFormatProperties(VkPhysicalDevice physicalDevice,
-                                        VkFormat format,
-                                        VkFormatProperties *pFormatProperties)
-{
-   VK_FROM_HANDLE(panvk_physical_device, physical_device, physicalDevice);
-
-   get_format_properties(physical_device, format, pFormatProperties);
-}
-
 VKAPI_ATTR void VKAPI_CALL
 panvk_GetPhysicalDeviceFormatProperties2(VkPhysicalDevice physicalDevice,
                                          VkFormat format,
@@ -1007,16 +1105,18 @@ panvk_GetPhysicalDeviceFormatProperties2(VkPhysicalDevice physicalDevice,
 
    VkDrmFormatModifierPropertiesListEXT *list = vk_find_struct(
       pFormatProperties->pNext, DRM_FORMAT_MODIFIER_PROPERTIES_LIST_EXT);
-   if (list) {
+   if (list && pFormatProperties->formatProperties.linearTilingFeatures) {
       VK_OUTARRAY_MAKE_TYPED(VkDrmFormatModifierPropertiesEXT, out,
                              list->pDrmFormatModifierProperties,
                              &list->drmFormatModifierCount);
 
-      vk_outarray_append_typed(VkDrmFormatModifierProperties2EXT, &out,
+      vk_outarray_append_typed(VkDrmFormatModifierPropertiesEXT, &out,
                                mod_props)
       {
          mod_props->drmFormatModifier = DRM_FORMAT_MOD_LINEAR;
          mod_props->drmFormatModifierPlaneCount = 1;
+         mod_props->drmFormatModifierTilingFeatures =
+            pFormatProperties->formatProperties.linearTilingFeatures;
       }
    }
 }
@@ -1041,25 +1141,30 @@ get_image_format_properties(struct panvk_physical_device *physical_device,
    case VK_IMAGE_TILING_LINEAR:
       format_feature_flags = format_props.linearTilingFeatures;
       break;
+   case VK_IMAGE_TILING_OPTIMAL:
+      format_feature_flags = format_props.optimalTilingFeatures;
+      break;
+   case VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT: {
+      const VkPhysicalDeviceImageDrmFormatModifierInfoEXT *mod_info =
+         vk_find_struct_const(
+            info->pNext, PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT);
+      if (mod_info->drmFormatModifier != DRM_FORMAT_MOD_LINEAR)
+         goto unsupported;
 
-   case VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT:
       /* The only difference between optimal and linear is currently whether
        * depth/stencil attachments are allowed on depth/stencil formats.
        * There's no reason to allow importing depth/stencil textures, so just
        * disallow it and then this annoying edge case goes away.
-       *
-       * TODO: If anyone cares, we could enable this by looking at the
-       * modifier and checking if it's LINEAR or not.
        */
       if (util_format_is_depth_or_stencil(format))
          goto unsupported;
 
       assert(format_props.optimalTilingFeatures ==
              format_props.linearTilingFeatures);
-      FALLTHROUGH;
-   case VK_IMAGE_TILING_OPTIMAL:
-      format_feature_flags = format_props.optimalTilingFeatures;
+
+      format_feature_flags = format_props.linearTilingFeatures;
       break;
+   }
    default:
       unreachable("bad VkPhysicalDeviceImageFormatInfo2");
    }
@@ -1067,32 +1172,28 @@ get_image_format_properties(struct panvk_physical_device *physical_device,
    if (format_feature_flags == 0)
       goto unsupported;
 
-   if (info->type != VK_IMAGE_TYPE_2D &&
-       util_format_is_depth_or_stencil(format))
-      goto unsupported;
-
    switch (info->type) {
    default:
       unreachable("bad vkimage type");
    case VK_IMAGE_TYPE_1D:
-      maxExtent.width = 16384;
+      maxExtent.width = 1 << 16;
       maxExtent.height = 1;
       maxExtent.depth = 1;
-      maxMipLevels = 15; /* log2(maxWidth) + 1 */
-      maxArraySize = 2048;
+      maxMipLevels = 17; /* log2(maxWidth) + 1 */
+      maxArraySize = 1 << 16;
       break;
    case VK_IMAGE_TYPE_2D:
-      maxExtent.width = 16384;
-      maxExtent.height = 16384;
+      maxExtent.width = 1 << 16;
+      maxExtent.height = 1 << 16;
       maxExtent.depth = 1;
-      maxMipLevels = 15; /* log2(maxWidth) + 1 */
-      maxArraySize = 2048;
+      maxMipLevels = 17; /* log2(maxWidth) + 1 */
+      maxArraySize = 1 << 16;
       break;
    case VK_IMAGE_TYPE_3D:
-      maxExtent.width = 2048;
-      maxExtent.height = 2048;
-      maxExtent.depth = 2048;
-      maxMipLevels = 12; /* log2(maxWidth) + 1 */
+      maxExtent.width = 1 << 16;
+      maxExtent.height = 1 << 16;
+      maxExtent.depth = 1 << 16;
+      maxMipLevels = 17; /* log2(maxWidth) + 1 */
       maxArraySize = 1;
       break;
    }
@@ -1107,28 +1208,41 @@ get_image_format_properties(struct panvk_physical_device *physical_device,
       sampleCounts |= VK_SAMPLE_COUNT_4_BIT;
    }
 
-   if (info->usage & VK_IMAGE_USAGE_SAMPLED_BIT) {
-      if (!(format_feature_flags & VK_FORMAT_FEATURE_SAMPLED_IMAGE_BIT)) {
-         goto unsupported;
+   /* From the Vulkan 1.2.199 spec:
+   *
+   *    "VK_IMAGE_CREATE_EXTENDED_USAGE_BIT specifies that the image can be
+   *    created with usage flags that are not supported for the format the
+   *    image is created with but are supported for at least one format a
+   *    VkImageView created from the image can have."
+   *
+   * If VK_IMAGE_CREATE_EXTENDED_USAGE_BIT is set, views can be created with
+   * different usage than the image so we can't always filter on usage.
+   * There is one exception to this below for storage.
+   */
+   if (!(info->flags & VK_IMAGE_CREATE_EXTENDED_USAGE_BIT)) {
+      if (info->usage & VK_IMAGE_USAGE_SAMPLED_BIT) {
+         if (!(format_feature_flags & VK_FORMAT_FEATURE_SAMPLED_IMAGE_BIT)) {
+            goto unsupported;
+         }
       }
-   }
 
-   if (info->usage & VK_IMAGE_USAGE_STORAGE_BIT) {
-      if (!(format_feature_flags & VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT)) {
-         goto unsupported;
+      if (info->usage & VK_IMAGE_USAGE_STORAGE_BIT) {
+         if (!(format_feature_flags & VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT)) {
+            goto unsupported;
+         }
       }
-   }
 
-   if (info->usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT) {
-      if (!(format_feature_flags & VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT)) {
-         goto unsupported;
+      if (info->usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT) {
+         if (!(format_feature_flags & VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT)) {
+            goto unsupported;
+         }
       }
-   }
 
-   if (info->usage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT) {
-      if (!(format_feature_flags &
-            VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT)) {
-         goto unsupported;
+      if (info->usage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT) {
+         if (!(format_feature_flags &
+               VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT)) {
+            goto unsupported;
+         }
       }
    }
 
@@ -1160,29 +1274,6 @@ unsupported:
    return VK_ERROR_FORMAT_NOT_SUPPORTED;
 }
 
-VKAPI_ATTR VkResult VKAPI_CALL
-panvk_GetPhysicalDeviceImageFormatProperties(
-   VkPhysicalDevice physicalDevice, VkFormat format, VkImageType type,
-   VkImageTiling tiling, VkImageUsageFlags usage,
-   VkImageCreateFlags createFlags,
-   VkImageFormatProperties *pImageFormatProperties)
-{
-   VK_FROM_HANDLE(panvk_physical_device, physical_device, physicalDevice);
-
-   const VkPhysicalDeviceImageFormatInfo2 info = {
-      .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2,
-      .pNext = NULL,
-      .format = format,
-      .type = type,
-      .tiling = tiling,
-      .usage = usage,
-      .flags = createFlags,
-   };
-
-   return get_image_format_properties(physical_device, &info,
-                                      pImageFormatProperties, NULL);
-}
-
 static VkResult
 panvk_get_external_image_format_properties(
    const struct panvk_physical_device *physical_device,
@@ -1190,50 +1281,51 @@ panvk_get_external_image_format_properties(
    VkExternalMemoryHandleTypeFlagBits handleType,
    VkExternalMemoryProperties *external_properties)
 {
-   VkExternalMemoryFeatureFlagBits flags = 0;
-   VkExternalMemoryHandleTypeFlags export_flags = 0;
-   VkExternalMemoryHandleTypeFlags compat_flags = 0;
+   const VkExternalMemoryHandleTypeFlags supported_handle_types =
+      VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT |
+      VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT;
 
-   /* From the Vulkan 1.1.98 spec:
-    *
-    *    If handleType is not compatible with the format, type, tiling,
-    *    usage, and flags specified in VkPhysicalDeviceImageFormatInfo2,
-    *    then vkGetPhysicalDeviceImageFormatProperties2 returns
-    *    VK_ERROR_FORMAT_NOT_SUPPORTED.
+   if (!(handleType & supported_handle_types)) {
+      return panvk_errorf(physical_device, VK_ERROR_FORMAT_NOT_SUPPORTED,
+                          "VkExternalMemoryTypeFlagBits(0x%x) unsupported",
+                          handleType);
+   }
+
+   /* pan_image_layout_init requires 2D for explicit layout */
+   if (pImageFormatInfo->type != VK_IMAGE_TYPE_2D) {
+      return panvk_errorf(
+         physical_device, VK_ERROR_FORMAT_NOT_SUPPORTED,
+         "VkExternalMemoryTypeFlagBits(0x%x) unsupported for VkImageType(%d)",
+         handleType, pImageFormatInfo->type);
+   }
+
+   /* There is no restriction on opaque fds.  But for dma-bufs, we want to
+    * make sure vkGetImageSubresourceLayout can be used to query the image
+    * layout of an exported dma-buf.  We also want to make sure
+    * VkImageDrmFormatModifierExplicitCreateInfoEXT can be used to specify the
+    * image layout of an imported dma-buf.  These add restrictions on the
+    * image tilings.
     */
-   switch (handleType) {
-   case VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT:
-   case VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT:
-      switch (pImageFormatInfo->type) {
-      case VK_IMAGE_TYPE_2D:
-         flags = VK_EXTERNAL_MEMORY_FEATURE_DEDICATED_ONLY_BIT |
-                 VK_EXTERNAL_MEMORY_FEATURE_EXPORTABLE_BIT |
-                 VK_EXTERNAL_MEMORY_FEATURE_IMPORTABLE_BIT;
-         compat_flags = export_flags =
-            VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT |
-            VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT;
-         break;
-      default:
-         return vk_errorf(
-            physical_device, VK_ERROR_FORMAT_NOT_SUPPORTED,
-            "VkExternalMemoryTypeFlagBits(0x%x) unsupported for VkImageType(%d)",
-            handleType, pImageFormatInfo->type);
-      }
-      break;
-   case VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT:
-      flags = VK_EXTERNAL_MEMORY_FEATURE_IMPORTABLE_BIT;
-      compat_flags = VK_EXTERNAL_MEMORY_HANDLE_TYPE_HOST_ALLOCATION_BIT_EXT;
-      break;
-   default:
-      return vk_errorf(physical_device, VK_ERROR_FORMAT_NOT_SUPPORTED,
-                       "VkExternalMemoryTypeFlagBits(0x%x) unsupported",
-                       handleType);
+   VkExternalMemoryFeatureFlags features = 0;
+   if (handleType == VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT ||
+       pImageFormatInfo->tiling == VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT) {
+      features |= VK_EXTERNAL_MEMORY_FEATURE_EXPORTABLE_BIT |
+                  VK_EXTERNAL_MEMORY_FEATURE_IMPORTABLE_BIT;
+   } else if (pImageFormatInfo->tiling == VK_IMAGE_TILING_LINEAR) {
+      features |= VK_EXTERNAL_MEMORY_FEATURE_EXPORTABLE_BIT;
+   }
+
+   if (!features) {
+      return panvk_errorf(
+         physical_device, VK_ERROR_FORMAT_NOT_SUPPORTED,
+         "VkExternalMemoryTypeFlagBits(0x%x) unsupported for VkImageTiling(%d)",
+         handleType, pImageFormatInfo->tiling);
    }
 
    *external_properties = (VkExternalMemoryProperties){
-      .externalMemoryFeatures = flags,
-      .exportFromImportedHandleTypes = export_flags,
-      .compatibleHandleTypes = compat_flags,
+      .externalMemoryFeatures = features,
+      .exportFromImportedHandleTypes = supported_handle_types,
+      .compatibleHandleTypes = supported_handle_types,
    };
 
    return VK_SUCCESS;
@@ -1310,6 +1402,11 @@ panvk_GetPhysicalDeviceImageFormatProperties2(
          &external_props->externalMemoryProperties);
       if (result != VK_SUCCESS)
          goto fail;
+
+      /* pan_image_layout_init requirements for explicit layout */
+      base_props->imageFormatProperties.maxMipLevels = 1;
+      base_props->imageFormatProperties.maxArrayLayers = 1;
+      base_props->imageFormatProperties.sampleCounts = 1;
    }
 
    if (cubic_props) {
@@ -1374,5 +1471,27 @@ panvk_GetPhysicalDeviceExternalBufferProperties(
    const VkPhysicalDeviceExternalBufferInfo *pExternalBufferInfo,
    VkExternalBufferProperties *pExternalBufferProperties)
 {
-   panvk_stub();
+   const VkExternalMemoryHandleTypeFlags supported_handle_types =
+      VK_EXTERNAL_MEMORY_HANDLE_TYPE_OPAQUE_FD_BIT |
+      VK_EXTERNAL_MEMORY_HANDLE_TYPE_DMA_BUF_BIT_EXT;
+
+   /* From the Vulkan 1.3.298 spec:
+    *
+    *    compatibleHandleTypes must include at least handleType.
+    */
+   VkExternalMemoryHandleTypeFlags handle_types =
+      pExternalBufferInfo->handleType;
+   VkExternalMemoryFeatureFlags features = 0;
+   if (pExternalBufferInfo->handleType & supported_handle_types) {
+      handle_types |= supported_handle_types;
+      features |= VK_EXTERNAL_MEMORY_FEATURE_EXPORTABLE_BIT |
+                  VK_EXTERNAL_MEMORY_FEATURE_IMPORTABLE_BIT;
+   }
+
+   pExternalBufferProperties->externalMemoryProperties =
+      (VkExternalMemoryProperties){
+         .externalMemoryFeatures = features,
+         .exportFromImportedHandleTypes = handle_types,
+         .compatibleHandleTypes = handle_types,
+      };
 }
